module SnowmanIO
  class App
    include Mongoid::Document
    include Mongoid::Timestamps
    include Concerns::Tokenable
    has_many :metrics, dependent: :destroy

    field :name,  type: String
    field :token, type: String
    field :system, type: Boolean, default: false

    validates :name, :token, presence: true

    before_validation on: :create do
      self.token = generate_token(:token)
    end

    def self.time(name, &block)
      app = App.where(system: true).first
      start = Time.now.to_f
      yield
      app.register_metric_value(name, "time", Time.now.to_f - start, Time.now)
    end

    def self.value(name, v, at = Time.now)
      app = App.where(system: true).first
      app.register_metric_value(name, "amount", v, at)
    end

    def as_json(options = {})
      super(options.merge(methods: [:datapoints, :metric_ids])).tap do |o|
        o["id"] = o.delete("_id").to_s
        o["metric_ids"] = o["metric_ids"].map(&:to_s)
      end
    end

    # Returns amount of datapoints for `at` and day before
    def daily_metrics(at)
      json = {}

      today = at.beginning_of_day
      yesterday = at.beginning_of_day - 1.day
      json["today"] = {"at" => today.strftime("%Y-%m-%d"), "count" => 0}
      json["yesterday"] = {"at" => yesterday.strftime("%Y-%m-%d"), "count" => 0}
      json["total"] = {"count" => 0}

      # TODO: impl it more performable
      metrics.each do |metric|
        if aggr = metric.aggregations.where(precision: "daily", at: today).first
          json["today"]["count"] += aggr.count.to_i
        end

        if aggr = metric.aggregations.where(precision: "daily", at: yesterday).first
          json["yesterday"]["count"] += aggr.count.to_i
        end

        json["total"]["count"] += metric.aggregations.where(precision: "daily").sum(:count).to_i
      end

      json
    end

    def register_metric_value(name, kind, value, at = Time.now)
      metric = metrics.where(name: name, kind: kind).first_or_create!
      metric.update_attributes!(last_value: value, last_value_updated_at: Time.now)
      metric.data_points.create!(at: at, value: value)
      touch
    end

    def datapoints
      daily_metrics(Time.now)
    end
  end
end
