module SnowmanIO
  class User
    include Mongoid::Document
    include Mongoid::Timestamps
    include ActiveModel::SecurePassword
    include Concerns::Tokenable

    has_secure_password
    has_many :checks, dependent: :destroy

    field :name,                 type: String
    field :email,                type: String
    field :password_digest,      type: String
    field :authentication_token, type: String
    field :invite_send_count,    type: Integer, default: 0
    field :invite_token,         type: String, default: ""
    field :status,               type: String, default: "active"
    field :restore_pass_token,   type: String

    validates :email, :authentication_token, presence: true
    validates :email, uniqueness: true

    has_many :following_joins, foreign_key: "user_id", class_name: "SnowmanIO::Following", dependent: :destroy
    has_many :follower_joins, foreign_key: "target_id", class_name: "SnowmanIO::Following", dependent: :destroy

    def followings
      following_joins.map(&:target)
    end

    def followers
      follower_joins.map(&:user)
    end

    def follow!(other)
      following_joins.find_or_create_by!(target_id: other.id)
    end

    def unfollow!(other)
      following_joins.where(target_id: other.id).destroy_all
    end

    def invite!(by)
      if self.invite_token.blank?
        self.invite_token = generate_token(:invite_token)
      end
      self.invite_send_count += 1
      save!

      SnowMailer.send_invite(self, Setting.get(SnowmanIO::BASE_URL_KEY), by).deliver_now
    end

    def restore_password!
      if self.restore_pass_token.blank?
        self.restore_pass_token = generate_token(:restore_pass_token)
      end
      save!

      SnowMailer.restore_password(self, Setting.get(SnowmanIO::BASE_URL_KEY)).deliver_now
    end

    before_validation on: :create do
      self.authentication_token = generate_token(:authentication_token)
    end

    def as_json(options = {})
      super(options.merge(except: ["authentication_token", "password_digest", "invite_token"])).tap { |o|
        o["id"] = o.delete("_id").to_s
        o["following_ids"] = followings.map(&:id).map(&:to_s)
        o["follower_ids"] = followers.map(&:id).map(&:to_s)
      }
    end
  end
end
